"""Functions for Trainer setups as in RLlib."""
# pylint:disable=missing-function-docstring,logging-format-interpolation
import copy
import logging
from typing import Any
from typing import Callable
from typing import Optional

from ray.rllib.agents.callbacks import DefaultCallbacks
from ray.rllib.agents.trainer import Trainer
from ray.rllib.env.normalize_actions import NormalizeActionWrapper
from ray.rllib.utils import merge_dicts
from ray.rllib.utils.deprecation import DEPRECATED_VALUE
from ray.rllib.utils.deprecation import deprecation_warning
from ray.rllib.utils.framework import try_import_tf
from ray.rllib.utils.from_config import from_config
from ray.tune.registry import _global_registry
from ray.tune.registry import ENV_CREATOR
from ray.tune.resources import Resources

logger = logging.getLogger(__name__)
tf = try_import_tf()


# ==============================================================================
# Trainer setup
# ==============================================================================
def make_env_creator(env_id: Optional[str], config: dict) -> Callable[[dict], Any]:
    if env_id:
        config["env"] = env_id
        # An already registered env.
        if _global_registry.contains(ENV_CREATOR, env_id):
            env_creator = _global_registry.get(ENV_CREATOR, env_id)
        # A class specifier.
        elif "." in env_id:

            def env_creator(env_config):
                return from_config(env_id, env_config)

        # Try gym.
        else:
            import gym  # soft dependency

            def env_creator(_):
                return gym.make(env_id)

    else:

        def env_creator(_):
            return

    if config["normalize_actions"]:
        env_creator = normalize_env_creator(env_creator)
    return env_creator


def check_and_resolve_framework_settings(config: dict):
    # Check and resolve DL framework settings.
    if "use_pytorch" in config and config["use_pytorch"] != DEPRECATED_VALUE:
        deprecation_warning("use_pytorch", "framework=torch", error=False)
        if config["use_pytorch"]:
            config["framework"] = "torch"
        config.pop("use_pytorch")
    if "eager" in config and config["eager"] != DEPRECATED_VALUE:
        deprecation_warning("eager", "framework=tfe", error=False)
        if config["eager"]:
            config["framework"] = "tfe"
        config.pop("eager")

    # Enable eager/tracing support.
    if tf and config["framework"] == "tfe":
        if not tf.executing_eagerly():
            tf.enable_eager_execution()
        logger.info(
            "Executing eagerly, with eager_tracing={}".format(config["eager_tracing"])
        )
    if tf and not tf.executing_eagerly() and config["framework"] != "torch":
        logger.info(
            "Tip: set framework=tfe or the --eager flag to enable "
            "TensorFlow eager execution"
        )


def normalize_env_creator(env_creator: callable) -> callable:
    inner = env_creator

    def normalize(env):
        import gym  # soft dependency

        if not isinstance(env, gym.Env):
            raise ValueError(
                "Cannot apply NormalizeActionActionWrapper to env of "
                "type {}, which does not subclass gym.Env.".format(type(env)),
            )
        return NormalizeActionWrapper(env)

    return lambda env_config: normalize(inner(env_config))


def validate_callbacks(config: dict) -> DefaultCallbacks:
    if not callable(config["callbacks"]):
        raise ValueError(
            "`callbacks` must be a callable method that "
            "returns a subclass of DefaultCallbacks, got {}".format(config["callbacks"])
        )
    return config["callbacks"]()


def set_rllib_log_level(config: dict):
    log_level = config.get("log_level")
    if log_level in ["WARN", "ERROR"]:
        logger.info(
            "Current log_level is {}. For more information, "
            "set 'log_level': 'INFO' / 'DEBUG' or use the -v and "
            "-vv flags.".format(log_level)
        )
    if config.get("log_level"):
        logging.getLogger("ray.rllib").setLevel(config["log_level"])


def setup_evaluation_config(config: dict) -> dict:
    # Update env_config with evaluation settings:
    extra_config = copy.deepcopy(config["evaluation_config"])
    # Assert that user has not unset "in_evaluation".
    assert "in_evaluation" not in extra_config or extra_config["in_evaluation"] is True
    extra_config.update(
        {
            "batch_mode": "complete_episodes",
            "rollout_fragment_length": 1,
            "in_evaluation": True,
        }
    )
    logger.debug("using evaluation_config: {}".format(extra_config))
    return merge_dicts(config, extra_config)


# ==============================================================================
# Trainer resource requests
# ==============================================================================
def default_resource_request(cls, config: dict) -> Resources:
    cf = dict(cls.options.defaults, **config)  # pylint:disable=invalid-name
    Trainer._validate_config(cf)  # pylint:disable=protected-access
    num_workers = cf["num_workers"] + cf["evaluation_num_workers"]
    return Resources(
        cpu=cf["num_cpus_for_driver"],
        gpu=cf["num_gpus"],
        memory=cf["memory"],
        object_store_memory=cf["object_store_memory"],
        extra_cpu=cf["num_cpus_per_worker"] * num_workers,
        extra_gpu=cf["num_gpus_per_worker"] * num_workers,
        extra_memory=cf["memory_per_worker"] * num_workers,
        extra_object_store_memory=cf["object_store_memory_per_worker"] * num_workers,
    )



"""Registry of modules for PyTorch policies."""
import torch.nn as nn
from gym.spaces import Space

from .ddpg import DDPG
from .mb_ddpg import ModelBasedDDPG
from .mb_sac import ModelBasedSAC
from .naf import NAF
from .sac import SAC
from .v0.catalog import get_module as get_module_v0

MODULES = {}


class RepeatedModuleNameError(Exception):
    """Exception raised for attempting to register a repeated module name.

    Args:
        cls: NN module class
    """

    def __init__(self, cls: type):
        super().__init__(f"Module class {cls.__name__} already in catalog")


def register(cls: type):
    """Register module class in catalog by class name.

    Adds module to global registry dict

    Args:
        cls: NN module class

    Raises:
        RepeatedModuleNameError: If `cls` is already in registry
    """
    try:
        MODULES[cls.__name__] = cls
    except KeyError:
        raise RepeatedModuleNameError(cls)

    return cls


def get_module(obs_space: Space, action_space: Space, config: dict) -> nn.Module:
    """Retrieve and construct module of given name.

    Args:
        obs_space: Observation space
        action_space: Action space
        config: Configurations for module construction and initialization
    """
    type_ = config.pop("type")
    if type_ not in MODULES:
        return get_module_v0(obs_space, action_space, {"type": type_, **config})

    cls = MODULES[type_]
    spec = cls.spec_cls.from_dict(config)
    return cls(obs_space, action_space, spec)


for _cls in (DDPG, NAF, SAC, ModelBasedDDPG, ModelBasedSAC):
    register(_cls)



"""Support for modules with stochastic policies."""
from typing import List

import gym.spaces as spaces
import torch
import torch.nn as nn
from ray.rllib.utils import override

import raylab.pytorch.nn as nnx
import raylab.pytorch.nn.distributions as ptd
from raylab.pytorch.nn.init import initialize_
from raylab.utils.dictionaries import deep_merge


def _build_fully_connected(obs_space, config):
    mlp = nnx.FullyConnected(in_features=obs_space.shape[0], **config["encoder"])
    mlp.apply(
        initialize_(
            activation=config["encoder"].get("activation"),
            **config["initializer_options"],
        )
    )
    return mlp


class StochasticActorMixin:
    """Adds constructor for modules with stochastic policies."""

    # pylint:disable=too-few-public-methods
    BASE_CONFIG = {
        "encoder": {"units": (32, 32), "activation": "Tanh"},
        "initializer_options": {"name": "xavier_uniform"},
        "input_dependent_scale": False,
    }

    @staticmethod
    def _make_actor(obs_space, action_space, config):
        config = deep_merge(
            StochasticActorMixin.BASE_CONFIG,
            config.get("actor", {}),
            False,
            ["encoder"],
        )

        if isinstance(action_space, spaces.Discrete):
            logits = _build_fully_connected(obs_space, config)
            params = nnx.CategoricalParams(logits.out_features, action_space.n)
            params_module = nn.Sequential(logits, params)
            dist_module = ptd.Categorical()
        elif isinstance(action_space, spaces.Box):
            logits = _build_fully_connected(obs_space, config)
            params = nnx.NormalParams(
                logits.out_features,
                action_space.shape[0],
                input_dependent_scale=config["input_dependent_scale"],
            )
            params_module = nn.Sequential(logits, params)
            dist_module = ptd.TransformedDistribution(
                ptd.Independent(ptd.Normal(), reinterpreted_batch_ndims=1),
                ptd.flows.TanhSquashTransform(
                    low=torch.as_tensor(action_space.low),
                    high=torch.as_tensor(action_space.high),
                    event_dim=1,
                ),
            )
        else:
            raise ValueError(f"Unsopported action space type {type(action_space)}")

        return {"actor": StochasticPolicy(params_module, dist_module)}


class MaximumEntropyMixin:
    """Adds entropy coefficient parameter to module."""

    # pylint:disable=too-few-public-methods
    BASE_CONFIG = {"initial_alpha": 1.0}

    def __init__(self, obs_space, action_space, config):
        super().__init__(obs_space, action_space, config)
        config = deep_merge(
            MaximumEntropyMixin.BASE_CONFIG, config.get("entropy", {}), False
        )
        self.alpha = Alpha(config["initial_alpha"])


class StochasticPolicy(nn.Module):
    """Represents a stochastic policy as a conditional distribution module."""

    # pylint:disable=abstract-method

    def __init__(self, params_module, dist_module):
        super().__init__()
        self.params = params_module
        self.dist = dist_module

    @override(nn.Module)
    def forward(self, obs):  # pylint:disable=arguments-differ
        return self.params(obs)

    @torch.jit.export
    def sample(self, obs, sample_shape: List[int] = ()):
        """
        Generates a sample_shape shaped sample or sample_shape shaped batch of
        samples if the distribution parameters are batched. Returns a (sample, log_prob)
        pair.
        """
        params = self(obs)
        return self.dist.sample(params, sample_shape)

    @torch.jit.export
    def rsample(self, obs, sample_shape: List[int] = ()):
        """
        Generates a sample_shape shaped reparameterized sample or sample_shape
        shaped batch of reparameterized samples if the distribution parameters
        are batched. Returns a (rsample, log_prob) pair.
        """
        params = self(obs)
        return self.dist.rsample(params, sample_shape)

    @torch.jit.export
    def log_prob(self, obs, action):
        """
        Returns the log of the probability density/mass function evaluated at `action`.
        """
        params = self(obs)
        return self.dist.log_prob(action, params)

    @torch.jit.export
    def cdf(self, obs, action):
        """Returns the cumulative density/mass function evaluated at `action`."""
        params = self(obs)
        return self.dist.cdf(action, params)

    @torch.jit.export
    def icdf(self, obs, prob):
        """Returns the inverse cumulative density/mass function evaluated at `prob`."""
        params = self(obs)
        return self.dist.icdf(prob, params)

    @torch.jit.export
    def entropy(self, obs):
        """Returns entropy of distribution."""
        params = self(obs)
        return self.dist.entropy(params)

    @torch.jit.export
    def perplexity(self, obs):
        """Returns perplexity of distribution."""
        params = self(obs)
        return self.dist.perplexity(params)

    @torch.jit.export
    def reproduce(self, obs, action):
        """Produce a reparametrized sample with the same value as `action`."""
        params = self(obs)
        return self.dist.reproduce(action, params)

    @torch.jit.export
    def deterministic(self, obs):
        """
        Generates a deterministic sample or batch of samples if the distribution
        parameters are batched. Returns a (rsample, log_prob) pair.
        """
        params = self(obs)
        return self.dist.deterministic(params)


class Alpha(nn.Module):
    # pylint:disable=missing-class-docstring

    def __init__(self, initial_alpha):
        super().__init__()
        self.log_alpha = nn.Parameter(torch.as_tensor(initial_alpha).log())

    def forward(self):  # pylint:disable=arguments-differ
        return self.log_alpha.exp()



# pylint:disable=missing-docstring
# pylint: enable=missing-docstring
import torch
import torch.distributions as dists


class DiagMultivariateNormal(dists.Independent):
    """
    Creates a multivariate Normal (also called Gaussian) distribution
    parameterized by a mean vector and a diagonal covariance matrix.
    """

    # pylint:disable=abstract-method

    def __init__(self, loc, scale_diag, validate_args=None):
        base_distribution = dists.Normal(loc=loc, scale=scale_diag)
        super().__init__(
            base_distribution, reinterpreted_batch_ndims=1, validate_args=validate_args
        )

    def reproduce(self, event):
        """Produce a reparametrized sample with the same value as `event`."""
        loc = self.base_dist.loc
        scale_diag = self.base_dist.scale
        with torch.no_grad():
            eps = (event - loc) / scale_diag
        return loc + scale_diag * eps



# pylint:disable=missing-docstring
# pylint: enable=missing-docstring
import torch.distributions as dists

from .transforms import TanhTransform


class SquashedDistribution(dists.TransformedDistribution):
    """Transformed distribution with samples constrained to the desired range."""

    # pylint:disable=abstract-method

    def __init__(self, base_distribution, low, high, **kwargs):
        squash = TanhTransform(cache_size=1)
        shift = dists.AffineTransform(
            loc=(high + low) / 2, scale=(high - low) / 2, cache_size=1, event_dim=1
        )
        super().__init__(base_distribution, [squash, shift], **kwargs)

    @property
    def mean(self):
        var = self.base_dist.mean
        for transform in self.transforms:
            var = transform(var)
        return var

    def reproduce(self, event):
        """Produce a reparametrized sample with the same value as `event`."""
        assert hasattr(
            self.base_dist, "reproduce"
        ), "Base distribution has no `reproduce` method"

        for transform in reversed(self.transforms):
            event = transform.inv(event)
        sample = self.base_dist.reproduce(event)
        for transform in self.transforms:
            sample = transform(sample)
        return sample



"""Modules mapping inputs to distribution parameters."""
from typing import Dict

import torch
import torch.nn as nn
from ray.rllib.utils import override

from raylab.pytorch.nn.init import initialize_

from .leaf_parameter import LeafParameter


class CategoricalParams(nn.Module):
    """Produce Categorical parameters.

    Initialized to be close to a discrete uniform distribution.
    """

    def __init__(self, in_features: int, n_categories: int):
        super().__init__()
        self.logits_module = nn.Linear(in_features, n_categories)
        self.apply(initialize_("orthogonal", gain=0.01))

    @override(nn.Module)
    def forward(self, inputs: torch.Tensor) -> Dict[str, torch.Tensor]:
        # pylint:disable=arguments-differ
        logits = self.logits_module(inputs)
        return {"logits": logits - logits.logsumexp(dim=-1, keepdim=True)}


class NormalParams(nn.Module):
    """Produce Normal parameters.

    Initialized to be close to a standard Normal distribution.
    """

    __constants__ = {"LOG_STD_MAX", "LOG_STD_MIN"}
    LOG_STD_MAX = 2
    LOG_STD_MIN = -20

    def __init__(
        self, in_features: int, event_size: int, input_dependent_scale: bool = True
    ):
        super().__init__()
        self.loc_module = nn.Linear(in_features, event_size)
        if input_dependent_scale:
            self.log_scale_module = nn.Linear(in_features, event_size)
        else:
            self.log_scale_module = LeafParameter(event_size)
        self.apply(initialize_("orthogonal", gain=0.01))

    @override(nn.Module)
    def forward(self, inputs: torch.Tensor) -> Dict[str, torch.Tensor]:
        # pylint:disable=arguments-differ
        loc = self.loc_module(inputs)
        log_scale = self.log_scale_module(inputs)
        scale = torch.clamp(log_scale, self.LOG_STD_MIN, self.LOG_STD_MAX).exp()
        return {"loc": loc, "scale": scale}


class StdNormalParams(nn.Module):
    """Produce Normal parameters with unit variance."""

    def __init__(self, input_dim: int, event_size: int):
        super().__init__()
        self.input_dim = input_dim
        self.event_shape = (event_size,)

    @override(nn.Module)
    def forward(self, inputs: torch.Tensor) -> Dict[str, torch.Tensor]:
        # pylint:disable=arguments-differ
        shape = inputs.shape[: -self.input_dim] + self.event_shape
        return {"loc": torch.zeros(shape), "scale": torch.ones(shape)}



"""Network and configurations for modules with deterministic policies."""
import warnings
from dataclasses import dataclass
from dataclasses import field

import torch.nn as nn
from dataclasses_json import DataClassJsonMixin
from gym.spaces import Box

from .policy.deterministic import DeterministicPolicy
from .policy.deterministic import MLPDeterministicPolicy

NetworkSpec = MLPDeterministicPolicy.spec_cls


@dataclass
class DeterministicActorSpec(DataClassJsonMixin):
    """Specifications for policy, behavior, and target policy.

    Args:
        network: Specifications for deterministic mlp policy network
        separate_behavior: Whether to create a separate behavior policy. Usually
            for parameter noise exploration, in which case it is recommended to
            enable encoder.layer_norm alongside this option.
        smooth_target_policy: Whether to use a noisy target policy for
            Q-Learning
        target_gaussian_sigma: Gaussian standard deviation for noisy target
            policy
        separate_target_policy: Whether to use separate parameters for the
            target policy. Intended for use with polyak averaging
        initializer: Optional dictionary with mandatory `type` key corresponding
            to the initializer function name in `torch.nn.init` and optional
            keyword arguments.
    """

    network: NetworkSpec = field(default_factory=NetworkSpec)
    separate_behavior: bool = False
    smooth_target_policy: bool = True
    target_gaussian_sigma: float = 0.3
    separate_target_policy: bool = False
    initializer: dict = field(default_factory=dict)

    def __post_init__(self):
        cls_name = type(self).__name__
        assert (
            self.target_gaussian_sigma > 0
        ), f"{cls_name}.target_gaussian_sigma must be positive"


class DeterministicActor(nn.Module):
    """NN with deterministic policies.

    Args:
        obs_space: Observation space
        action_space: Action space
        spec: Specifications for policy, behavior, and target policy

    Attributes:
        policy: The deterministic policy to be learned
        behavior: The policy for exploration. `utils.exploration.GaussianNoise`
            handles Gaussian action noise exploration separatedly.
        target_policy: The policy used for estimating the arg max in Q-Learning
        spec_cls: Expected class of `spec` init argument
    """

    # pylint:disable=abstract-method
    spec_cls = DeterministicActorSpec

    def __init__(
        self, obs_space: Box, action_space: Box, spec: DeterministicActorSpec,
    ):
        super().__init__()

        def make_policy():
            return MLPDeterministicPolicy(obs_space, action_space, spec.network)

        policy = make_policy()
        policy.initialize_parameters(spec.initializer)

        behavior = policy
        if spec.separate_behavior:
            if not spec.network.layer_norm:
                warnings.warn(
                    "Separate behavior policy requested and layer normalization"
                    " deactivated. If using parameter noise exploration, enable"
                    " layer normalization for better stability."
                )
            behavior = make_policy()
            behavior.load_state_dict(policy.state_dict())

        target_policy = policy
        if spec.separate_target_policy:
            target_policy = make_policy()
            target_policy.load_state_dict(policy.state_dict())
        if spec.smooth_target_policy:
            target_policy = DeterministicPolicy.add_gaussian_noise(
                target_policy, noise_stddev=spec.target_gaussian_sigma
            )

        self.policy = policy
        self.behavior = behavior
        self.target_policy = target_policy



"""Policy for MBPO using PyTorch."""
from raylab.agents.sac import SACTorchPolicy
from raylab.policy import EnvFnMixin
from raylab.policy import ModelSamplingMixin
from raylab.policy import ModelTrainingMixin
from raylab.policy.action_dist import WrapStochasticPolicy
from raylab.policy.losses import ModelEnsembleMLE
from raylab.pytorch.optim import build_optimizer


class MBPOTorchPolicy(
    EnvFnMixin, ModelTrainingMixin, ModelSamplingMixin, SACTorchPolicy
):
    """Model-Based Policy Optimization policy in PyTorch to use with RLlib."""

    # pylint:disable=too-many-ancestors
    dist_class = WrapStochasticPolicy

    def __init__(self, observation_space, action_space, config):
        super().__init__(observation_space, action_space, config)
        models = self.module.models
        self.loss_model = ModelEnsembleMLE(models)

    @property
    def options(self):
        # pylint:disable=cyclic-import
        from raylab.agents.mbpo import MBPOTrainer

        return MBPOTrainer.options

    @property
    def model_training_loss(self):
        return self.loss_model

    def _make_optimizers(self):
        optimizers = super()._make_optimizers()
        config = self.config["torch_optimizer"]
        optimizers["models"] = build_optimizer(self.module.models, config["models"])
        return optimizers



"""Generic Trainer and base configuration for model-based agents."""
import logging
from typing import Any
from typing import List
from typing import Tuple

from ray.rllib import Policy
from ray.rllib import SampleBatch
from ray.rllib.evaluation.worker_set import WorkerSet
from ray.rllib.utils import override
from ray.rllib.utils.timer import TimerStat

from raylab.agents import trainer
from raylab.agents.off_policy import OffPolicyTrainer
from raylab.utils.annotations import StatDict
from raylab.utils.replay_buffer import NumpyReplayBuffer


logger = logging.getLogger(__name__)


def set_policy_with_env_fn(worker_set: WorkerSet, fn_type: str = "reward"):
    """Set the desired environment function for all policies in the worker set.

    Args:
        worker_set: A worker set instance, usually from a trainer
        fn_type: The type of environment function, either 'reward',
            'termination', or 'dynamics'
        from_env: Whether to retrieve the function from the environment instance
            or from the global registry
    """
    worker_set.foreach_worker(
        lambda w: w.foreach_policy(
            lambda p, _: _set_from_env_if_possible(p, w.env, fn_type)
        )
    )


def _set_from_env_if_possible(policy: Policy, env: Any, fn_type: str = "reward"):
    env_fn = getattr(env, fn_type + "_fn", None)
    if fn_type == "reward":
        if env_fn:
            policy.set_reward_from_callable(env_fn)
        else:
            policy.set_reward_from_config()
    elif fn_type == "termination":
        if env_fn:
            policy.set_termination_from_callable(env_fn)
        else:
            policy.set_termination_from_config()
    elif fn_type == "dynamics":
        if env_fn:
            policy.set_dynamics_from_callable(env_fn)
        else:
            raise ValueError(
                f"Environment '{env}' has no '{fn_type + '_fn'}' attribute"
            )
    else:
        raise ValueError(f"Invalid env function type '{fn_type}'")


@trainer.configure
@trainer.option(
    "holdout_ratio",
    default=0.2,
    help="Fraction of replay buffer to use as validation dataset"
    " (hence not for training)",
)
@trainer.option(
    "max_holdout",
    default=5000,
    help="Maximum number of samples to use as validation dataset",
)
@trainer.option(
    "model_update_interval",
    default=1,
    help="""Number of calls to rollout worker between each model update run.

    Will collect this many rollout fragments of length 'rollout_fragment_length'
    between calls to `train_dynamics_model`

    Example:
        With a 'rollout_fragment_length' of 1 and 'model_update_interval' of 25,
        will collect 25 environment transitions between each model optimization
        loop.
    """,
)
@trainer.option(
    "policy_improvement_interval",
    default=1,
    help="""Number of rollout worker calls between each `policy.learn_on_batch` call.

    Uses the same semantics as 'model_update_interval'.
    """,
)
class ModelBasedTrainer(OffPolicyTrainer):
    """Generic trainer for model-based agents.

    Sets reward and termination functions for policies. These functions must be
    either:
    * Registered via `raylab.envs.register_reward_fn` and
      `raylab.envs.register_termination_fn`
    * Accessible attributes of the environment as `reward_fn` and
      `termination_fn`. These should not be bound instance methods; all
      necessary information should be encoded in the inputs, (state, action,
      and next state) i.e., the states should be markovian.

    Policies must implement `optimize_model` according to
    `raylab.policy:ModelTrainingMixin`
    """

    # pylint:disable=attribute-defined-outside-init

    @override(OffPolicyTrainer)
    def _init(self, config, env_creator):
        super()._init(config, env_creator)
        self.model_timer = TimerStat()
        self.policy_timer = TimerStat()
        set_policy_with_env_fn(self.workers, fn_type="reward")
        set_policy_with_env_fn(self.workers, fn_type="termination")

        self._sample_calls: int = 0

    @staticmethod
    @override(OffPolicyTrainer)
    def validate_config(config):
        OffPolicyTrainer.validate_config(config)
        assert (
            config["holdout_ratio"] < 1.0
        ), "Holdout data cannot be the entire dataset"
        assert (
            config["max_holdout"] >= 0
        ), "Maximum number of holdout samples must be non-negative"
        assert (
            config["policy_improvements"] >= 0
        ), "Number of policy improvement steps must be non-negative"

    @override(OffPolicyTrainer)
    def _train(self):
        pre_learning_steps = self.sample_until_learning_starts()
        if pre_learning_steps:
            logger.info("Starting model warmup")
            _, warmup_stats = self.train_dynamics_model(warmup=True)
            # pylint:disable=logging-too-many-args
            logger.info("Finished model warmup with stats: %s", warmup_stats)

        timesteps_this_iter = 0
        config = self.config
        worker = self.workers.local_worker()
        stats = {}

        while timesteps_this_iter < max(config["timesteps_per_iteration"], 1):
            samples = worker.sample()
            self._sample_calls += 1
            timesteps_this_iter += samples.count
            for row in samples.rows():
                self.replay.add(row)

            if self._sample_calls % config["model_update_interval"] == 0:
                with self.model_timer:
                    _, model_info = self.train_dynamics_model()
                    self.model_timer.push_units_processed(model_info["model_epochs"])
                stats.update(model_info)

            if self._sample_calls % config["policy_improvement_interval"] == 0:
                with self.policy_timer:
                    policy_info = self.improve_policy(
                        times=config["policy_improvements"]
                    )
                    self.policy_timer.push_units_processed(
                        config["policy_improvements"]
                    )
                stats.update(policy_info)

        self.metrics.num_steps_sampled += timesteps_this_iter
        return self._log_metrics(stats, timesteps_this_iter + pre_learning_steps)

    @override(OffPolicyTrainer)
    def _log_metrics(self, learner_stats, timesteps_this_iter):
        metrics = super()._log_metrics(learner_stats, timesteps_this_iter)
        metrics.update(
            model_time_s=round(self.model_timer.mean, 3),
            policy_time_s=round(self.policy_timer.mean, 3),
            # Get mean number of model epochs per second spent updating the model
            model_update_throughput=round(self.model_timer.mean_throughput, 3),
            # Get mean number of policy updates per second spent updating the policy
            policy_update_throughput=round(self.policy_timer.mean_throughput, 3),
        )
        return metrics

    def train_dynamics_model(self, warmup: bool = True) -> Tuple[List[float], StatDict]:
        """Implements the model training step.

        Calls the policy to optimize the model on the environment replay buffer.

        Args:
            warmup: Whether the optimization is being done on data collected
                via :meth:`sample_until_learning_starts`.

        Returns:
            A tuple containing the list of evaluation losses for each model and
            a dictionary of training statistics
        """
        samples = self.replay.all_samples()
        samples.shuffle()
        holdout = min(
            int(len(self.replay) * self.config["holdout_ratio"]),
            self.config["max_holdout"],
        )
        train_data, eval_data = samples.slice(holdout, None), samples.slice(0, holdout)
        eval_data = None if eval_data.count == 0 else eval_data

        policy = self.get_policy()
        eval_losses, stats = policy.optimize_model(train_data, eval_data, warmup=warmup)

        return eval_losses, stats

    def improve_policy(self, times: int) -> StatDict:
        """Improve the policy on previously collected environment data.

        Calls the policy to learn on batches samples from the replay buffer.

        Args:
            times: number of times to call `policy.learn_on_batch`

        Returns:
            A dictionary of training and exploration statistics
        """
        policy = self.get_policy()
        batch_size = self.config["train_batch_size"]

        stats = {}
        for _ in range(times):
            batch = self.replay.sample(batch_size)
            stats.update(policy.learn_on_batch(batch))
            self.metrics.num_steps_trained += batch.count

        stats.update(policy.get_exploration_info())
        return stats


@trainer.configure
@trainer.option(
    "virtual_buffer_size",
    default=int(1e6),
    help="Size of the buffer for virtual samples",
)
@trainer.option(
    "model_rollouts",
    default=40,
    help="""Number of model rollouts to add to virtual buffer each policy interval.

    Populates virtual replay with this many model rollouts before each policy
    improvement.
    """,
)
@trainer.option(
    "real_data_ratio",
    default=0.1,
    help="Fraction of each policy minibatch to sample from environment replay pool",
)
class DynaLikeTrainer(ModelBasedTrainer):
    """Generic trainer for model-based agents with dyna-like data augmentation.

    If `model_rollouts` > 0, policies must implement `setup_sampling_models`
    and `generate_virtual_sample_batch` according to
    `raylab.policy:ModelSamplingMixin`
    """

    # pylint:disable=attribute-defined-outside-init

    @staticmethod
    def validate_config(config):
        ModelBasedTrainer.validate_config(config)
        assert (
            0 <= config["real_data_ratio"] <= 1
        ), "Fraction of real data samples for policy improvement must be in [0, 1]"
        assert (
            config["virtual_buffer_size"] >= 0
        ), "Virtual buffer capacity must be non-negative"
        assert (
            config["model_rollouts"] >= 0
        ), "Cannot sample a negative number of model rollouts"

    @override(OffPolicyTrainer)
    def build_replay_buffer(self, config):
        super().build_replay_buffer(config)
        policy = self.get_policy()
        self.virtual_replay = NumpyReplayBuffer(
            policy.observation_space, policy.action_space, config["virtual_buffer_size"]
        )
        self.virtual_replay.seed(config["seed"])

    @override(ModelBasedTrainer)
    def _train(self):
        pre_learning_steps = self.sample_until_learning_starts()
        timesteps_this_iter = 0

        config = self.config
        worker = self.workers.local_worker()
        policy = self.get_policy()
        stats = {}
        while timesteps_this_iter < max(config["timesteps_per_iteration"], 1):
            samples = worker.sample()
            self._sample_calls += 1
            timesteps_this_iter += samples.count
            for row in samples.rows():
                self.replay.add(row)

            if self._sample_calls % config["model_update_interval"] == 0:
                with self.model_timer:
                    eval_losses, model_info = self.train_dynamics_model()
                    self.model_timer.push_units_processed(model_info["model_epochs"])
                policy.set_new_elite(eval_losses)
                stats.update(model_info)

            if self._sample_calls % config["policy_improvement_interval"] == 0:
                self.populate_virtual_buffer(config["model_rollouts"])
                with self.policy_timer:
                    policy_info = self.improve_policy(
                        times=config["policy_improvements"]
                    )
                    self.policy_timer.push_units_processed(
                        config["policy_improvements"]
                    )
                stats.update(policy_info)

        self.metrics.num_steps_sampled += timesteps_this_iter
        return self._log_metrics(stats, timesteps_this_iter + pre_learning_steps)

    def populate_virtual_buffer(self, num_rollouts: int):
        """Add model rollouts branched from real data to the virtual pool.

        Args:
            num_rollouts: number of initial states to samples from the
                environment replay buffer
        """
        if not (num_rollouts and self.config["real_data_ratio"] < 1.0):
            return

        real_samples = self.replay.sample(num_rollouts)
        policy = self.get_policy()
        virtual_samples = policy.generate_virtual_sample_batch(real_samples)
        for row in virtual_samples.rows():
            self.virtual_replay.add(row)

    def improve_policy(self, times: int) -> StatDict:
        """Improve the policy on a mixture of environment and model data.

        Calls the policy to learn on batches sampled from the environment and
        model rollouts.

        Args:
            times: number of times to call `policy.learn_on_batch`

        Returns:
            A dictionary of training and exploration statistics
        """
        policy = self.get_policy()
        batch_size = self.config["train_batch_size"]
        env_batch_size = int(batch_size * self.config["real_data_ratio"])
        model_batch_size = batch_size - env_batch_size

        stats = {}
        for _ in range(times):
            samples = []
            if env_batch_size:
                samples += [self.replay.sample(env_batch_size)]
            if model_batch_size:
                samples += [self.virtual_replay.sample(model_batch_size)]
            batch = SampleBatch.concat_samples(samples)
            stats.update(policy.learn_on_batch(batch))
            self.metrics.num_steps_trained += batch.count

        stats.update(policy.get_exploration_info())
        return stats



# pylint:disable=missing-module-docstring
from abc import ABC
from abc import abstractmethod
from typing import List

import torch
import torch.nn as nn
from gym.spaces import Box
from torch import Tensor

from raylab.policy.modules.actor.policy.deterministic import DeterministicPolicy
from raylab.policy.modules.networks.mlp import StateMLP

from .q_value import QValue


MLPSpec = StateMLP.spec_cls


class VValue(ABC, nn.Module):
    """Neural network module emulating a V value function."""

    # pylint:disable=arguments-differ
    @abstractmethod
    def forward(self, obs: Tensor) -> Tensor:
        """Main forward pass mapping obs to V-values.

        Note:
            The output tensor has a last singleton dimension, i.e., for a batch
            of 10 observations, the output will have shape (10, 1).
        """


class PolicyQValue(VValue):
    """State-value function from policy and Q-value function.

    Args:
        policy: Deterministic policy
        q_value: Q-value function
    """

    def __init__(self, policy: DeterministicPolicy, q_value: QValue):
        super().__init__()
        self.policy = policy
        self.q_value = q_value

    def forward(self, obs: Tensor) -> Tensor:
        act = self.policy(obs)
        value = self.q_value(obs, act)
        return value


class MLPVValue(VValue):
    """V-value function with a multilayer perceptron encoder.

    Args:
        obs_space: Observation space
        spec: Multilayer perceptron specifications

    Attributes:
        encoder: NN module mapping observations to logits
        value_linear: Linear module mapping logits to values
    """

    spec_cls = MLPSpec

    def __init__(self, obs_space: Box, spec: MLPSpec):
        super().__init__()

        self.encoder = StateMLP(obs_space, spec)
        self.value_linear = nn.Linear(self.encoder.out_features, 1)

    def forward(self, obs: Tensor) -> Tensor:
        logits = self.encoder(obs)
        return self.value_linear(logits)


class VValueEnsemble(nn.ModuleList, VValue):
    """A static list of V-value estimators.

    Args:
        v_values: A list of VValue modules
    """

    def __init__(self, v_values: List[VValue]):
        cls_name = type(self).__name__
        assert all(
            isinstance(v, VValue) for v in v_values
        ), f"All modules in {cls_name} must be instances of VValue."
        super().__init__(v_values)

    def forward(self, obs: Tensor) -> Tensor:
        """Evaluate each V estimator in the ensemble.

        Args:
            obs: The observation tensor

        Returns:
            A tensor of shape `(*, N)`, where `N` is the ensemble size
        """
        # pylint:disable=arguments-differ
        return self._state_values(obs)

    def _state_values(self, obs: Tensor) -> Tensor:
        return torch.cat([m(obs) for m in self], dim=-1)


class ForkedVValueEnsemble(VValueEnsemble):
    """Ensemble of V-value estimators with parallelized forward pass."""

    def _state_values(self, obs: Tensor) -> Tensor:
        # pylint:disable=protected-access
        futures = [torch.jit._fork(m, obs) for m in self]
        return torch.cat([torch.jit._wait(f) for f in futures], dim=-1)
